function [A,R,P]=MODGS(A,R,P)
% [A,R,P]=MODGS(A)
% [A,R,P]=MODGS(A,R,P)
%
% sparse QR decomposition AP=QR using modified Gram-Schmidt. A is overwritten by
% Q. column pivoting is performed with respect to the number of nonzeros for
% any column of A. R is stored seperately.
% If R is input of size lxl, then only the columns  A(:,l+1:end) will be
% orthogonalized against the previous ones. If we already computed a QR 
% decomposition of A(:1:l) we can continue this decomposition for the final
% columns of A.
% Since A is overwritten, we finally have A_old(:,P)=A_new(:,P)*R
%
% ------------------------------------------------------------------------------
%
% Input
% -----
% A       Matrix whose columns have to be orthogonalized
% R       OPTIONAL, size lxl. Refers to R-part of QR decomposition of A(:,1:l)
%         In this case A(:,1:l) refers to the Q-part.
% P       OPTIONAL. permutation vector to a priori reduce the fill-in, extended
%         on output
%
% ------------------------------------------------------------------------------
%
% Output
% ------
% A       A'*A=I. Overwritten by the Q-part of the QR decomposition
% R       Upper triangular. R-part of the QR-decomposition
% P       permutation vector to a priori reduce the fill-in


[m,n]=size(A);

if nargin==1
   startpos=1;
   P=[];
else
   startpos=size(R,2)+1;
end;
P=[P startpos:n];

% reorder columns of A according to fill-in
% This is not the best choice, but for relatively small sparse matrices okay.
for j=startpos:n
    v(j-startpos+1)=nnz(A(:,j));
end % for j
[v,I]=sort(v);
P(startpos:n)=P(I+startpos-1);

for j=startpos:n
    
    % from the actual vector subtract the orthogonal projection to its 
    % predecessors
    if (j>1)
       % Compute projection (I-Q(:,1:j-1)*Q(:,1:j-1)')*A(:,j)       
       % In theory this is equivalent to
       % (I-Q(:,j-1)*Q(:,j-1)')*...*(I-Q(:,1)*Q(:,1)')*A(:,j)       
       for k=1:j-1 
	   R(k,j)=A(:,P(k))'*A(:,P(j));
           A(:,P(j))=A(:,P(j))-A(:,P(k))*R(k,j);
       end % for k
    end % if

    % Compute 2-Norm of the new orthogonalized vector
    R(j,j)=norm(A(:,P(j)),2);
    A(:,P(j))=A(:,P(j))/R(j,j);
end; % for j


