function [Y,PARAM]=PGMRES(alpha,b,Y,PARAM)
% y=PGMRES(alpha,x,y,PARAM)
% [y,PARAM]=PGMRES(alpha,x,y,PARAM)
% 
% (right) precondiioned GMRES iteration for the solution of a linear system. 
% y=alpha*PGMRES(x)+y
% 
% -----------------------------------------------------------------------------
%
% Input
% -----
% alpha             scalar
% x                 input matrix
% y                 given approximate solution
% 
% PARAM             structure containing additional parameters
% PARAM.A           system matrix matrix or function string for matrix vector 
%                   multiplication
% PARAM.APARAM      list of parameters if PARAM.A is a function string
%                   [z,APARAM]=A(alpha,x,y,APARAM)
% PARAM.P           approximate inverse to A or function string
% PARAM.PPARAM      list of parameters if PARAM.P is a function string
%                   [z,PPARAM]=P(alpha,x,y,PPARAM)
% PARAM.maxit       maximum number of steps, default inf
% PARAM.tol         relative tolerance for termination of CG default sqrt(eps)
% PARAM.x0          OPTIONAL, initial solution. default 0
% PARAM.restart     OPTIONAL, restart length, default n
% 
% -------------------------------------------------------------------------
%
% Output
% ------
% z                 solution
% PARAM             not altered
% PARAM.it          number of iterations
% PARAM.tol         set to its default value if not present on input


n=size(b,1);

% number of right hand sides
% nrhs=size(b,2);

% set tol to its default value if not prescribed
if ~isfield(PARAM,'tol')
   PARAM.tol=sqrt(eps);
end % if

% set maxit to its default value if not prescribed
if ~isfield(PARAM,'maxit')
   PARAM.maxit=inf;
end % if

% set restart to its default value if not prescribed
if ~isfield(PARAM,'restart')
   PARAM.restart=n;
end % if
PARAM.it=0; 

% ---------------    Some general internal preparations   ------------------
% restart length
% Select whether the residual should be printed during the
% iteration or not
PRINT_RES=0;
% decide whether the true residual and graphic information should be
% displayed or not
stat=0;
      
% decide whether SVD should be used to solve the system with H or not
% if USE_SVD==0, then it may happen that a restart has to be 
% performed earlier than after l steps. This will be the case if H
% has no longer full rank.
USE_SVD=0;
% information is printed out?
PRINT_INFO=0;

% Select choice of orthogonalization in GMRES
% Case 0: use modified Gram-Schmidt  (standard, but hard to 
%         parallelize)
% Case 1: use classical Gram-Schmidt (easiest way to parallelize, but
%                                     possibly loss of orthogonality)
% Case 2: use Householder            (most stable version, but also 
%                                     most expensive and hard to 
%                                     parallelize)
% Case 3: use modified Householder   (stable? version, as expensive 
%	                                as Householder but easy to 
%                                     parallelize)
CHOICE_ORTH=0;
if PRINT_INFO
if CHOICE_ORTH==0
   fprintf('\nUse MODIFIED Gram-Schmidt in GMRES\n');
elseif CHOICE_ORTH==1
   fprintf('\nUse CLASSICAL Gram-Schmidt in GMRES\n');
elseif CHOICE_ORTH==2   
   fprintf('\nUse Householder in GMRES\n');
else      
   fprintf('\nUse MODIFIED Householder in GMRES\n');
end % if-elseif-else
end % if PRINT_INFO
      
if stat
   clf;
end % if
% --------------   END Some general internal preparations   ---------------

      
      
% --------------------- GMRES iteration loop ---------------------
% Compute initial residual  = b-A*x0
% if there exists x0, set r=b-Ax0, else use r=b
if isfield(PARAM,'x0')
   x=PARAM.x0;
   if isstr(PARAM.A)
      [r,PARAM.APARAM]=feval(PARAM.A,-1,x,b,PARAM.APARAM);
   else
      r=b-PARAM.A*x;
   end % if-else
else
   x=zeros(n,1);
   r=b;   
end


% vectors containing residual information
res=[]; 

PARAM.restart=min(n,PARAM.restart);
y=zeros(n,1);
% norm of the residual
res=zeros(1,fix(1.5*n+PARAM.restart));
res2=zeros(1,fix(1.5*n+PARAM.restart));
% matrix with Arnoldi vectors AV=VH+...
V=zeros(n,PARAM.restart+1);
% small Hessenberg matrix
H=zeros(PARAM.restart+2,PARAM.restart+1);
% additional parameters for Householder representation of V
if CHOICE_ORTH>=2 % -------------------
beta=zeros(1,PARAM.restart+1);
if CHOICE_ORTH==3 % -------------------
T=zeros(PARAM.restart+1,PARAM.restart+1);
end % if CHOICE_ORTH==3 --------------
end % if CHOICE_ORTH>=2 ---------------
% auxilaury vector
Q=zeros(n,1);
% Parameters for Givens rotations
c=zeros(1,PARAM.restart+1);
s=zeros(1,PARAM.restart+1);   
   
       
% print residual information       
if PRINT_RES %%%%%%%%%%%%%%%%%%
fprintf('2-norm of the initial residual: %12.4e\n', norm(r,2));
clf;
end % if PRINT_RES %%%%%%%%%%%%%
  
% counter for total number of iterations
kk=1;

% print residual information       
if PRINT_RES % ----------------
fprintf('step %3d, ', kk);
end % if PRINT_RES -----------

% store history of the residual 
res(kk)=norm(r,2);

while (abs(res(kk))>PARAM.tol*abs(res(1)) & kk<=PARAM.maxit)
      if (kk>1)
	 % Compute the residual r = b-A*x 
	 if isstr(PARAM.A)
	    [r,PARAM.APARAM]=feval(PARAM.A,-1,x,b,PARAM.APARAM);
	 else
	    r=b-PARAM.A*x;
	 end % if-else
	 res(kk)=norm(r,2);
      end % if

      % first Arnoldi vector
      V(:,1)=r;
	     
% Choice of reorthogonalization scheme	     
if CHOICE_ORTH<2 % ------------------
      V(:,1)=V(:,1)/res(kk);
else % CHOICE_ORTH>=2 
      % compute Householder representation of r
      [V(:,1),beta(1),MU]=HOUSE(V(:,1));
      % Applying the Householder transformation to r will give a 
      % multiple of the fiPARAM.restart unit vector. This multiple will be
      % -sign(r(1))*MU.
      if (r(1)>=0)
	 res(kk)=-MU;
      else   
	 res(kk)=MU;
      end % if-else
if CHOICE_ORTH==3 % -------------------
      T(1,1)=1/beta(1);
end % if CHOICE_ORTH==3 --------------
end % if CHOICE_ORTH<2 ---------------
	 
if PRINT_RES % -----------------------
      if (kk>1)
	 fprintf('\nstep %3d, ', kk);
      end % if
      fprintf('residual%16.8e', abs(res(kk))); 
      semilogy(1:kk,abs(res(1:kk)),'-'); hold on;
      fprintf('\n'); 
end % if PRINT_RES ------------------
        

      i=1;
      % rhs is a vector which is used as monitor to check the 
      % residual. Using rhs, one can avoid the update of x in all 
      % steps until the iteration is terminated. 
      % Algebraically we have rhs=V'*r and later rhs=G*V'*r, where G
      % is the matrix with Givens rotations, which transforms H to
      % upper triangular form.
      rhs(1)=res(kk);
      % flag for nonsingular transformed Hessenberg matrix
      nsgl=1;
      % control vector for ||H||
      nrmH=zeros(PARAM.restart+2,1);
      % during the iteration the preconditioned residual will be 
      % estimated. Ensure that the pseudo residual is small enough 
      while (i<=PARAM.restart & abs(res(kk))>PARAM.tol*abs(res(1)) & nsgl & kk<=PARAM.maxit)
            if (kk>1 & i==1)
if PRINT_RES % ----------------
               fprintf('RESTART GMRES\n');
end % if PRINT_RES -----------
	    end % if

		      
	    % Q=A*M^{-1}*V(:,i);
if CHOICE_ORTH<2 % ------------------
	    Q=V(:,i);
elseif CHOICE_ORTH==2 
            % Compute V(:,i) from its Householder representation 
	    % HH_1*...*HH_i*e_i
	    Q=zeros(n,1);
	    Q(i)=1;
	    for k=i:-1:1
	        Q(k:n)=ROWHOUSE(Q(k:n),V(k:n,k),beta(k));
	    end % for
else % CHOICE_ORTH==3
            % Compute V(:,i) from its modified Householder representation 
	    % (I + V(:,1:i) T^{-1} V(:,1:i)') * e_i
	    Q=zeros(n,1);
	    Q(i)=1;
	    Q=Q+V(:,1:i)*(T(1:i,1:i)\V(i,1:i)');
end % if ---------------------------
	    % ------------------------------------------------------
	    % Solve the preconditioned system M*Q=Q	      
	    if isstr(PARAM.P)
	       [Q,PARAM.PPARAM]=feval(PARAM.P,1,Q,zeros(n,1),PARAM.PPARAM);
	    else
	       Q=PARAM.P*Q;
	    end % if-else
	    % -------------------------------------------------------
	    % Q = A * (M^{-1}*Q)
	    if isstr(PARAM.A)
	       [Q,PARAM.APARAM]=feval(PARAM.A,1,Q,zeros(n,1),PARAM.APARAM);
	    else
	       Q=PARAM.A*Q;
	    end % if-else

		      
	    % Compute next column of H 
if CHOICE_ORTH==0 % ----------------------
            % perform modified Gram-Schmidt orthogonalization
	    for k=1:i
	        H(k,i)=V(:,k)'*Q;
		Q     =Q-V(:,k)*H(k,i);
	    end % for
	    H(i+1,i)=norm(Q,2);
	    V(:,i+1)=Q/H(i+1,i);
elseif CHOICE_ORTH==1 % ----------------------
            % perform classical Gram-Schmidt orthogonalization
	    H(1:i,i)=V(:,1:i)'*Q;
	    Q       =Q-V(:,1:i)*H(1:i,i);
	    H(i+1,i)=norm(Q,2);
	    V(:,i+1)=Q/H(i+1,i);
elseif CHOICE_ORTH==2 % ------------------
            % Compute Householder representation of V(:,i+1)
	    % Step 1: Multiply with previous Householder transformations,
	    %         V(:,i+1) = HH_i*...*HH_1*Q
	    V(:,i+1)=Q;
	    for k=1:i
	        V(k:n,i+1)=ROWHOUSE(V(k:n,i+1),V(k:n,k),beta(k));
	    end % for
	    % => H(1:i,i)=V(:,1:i)'*Q=V(1:i,i+1)
	    H(1:min(n,i+1),i)=V(1:min(n,i+1),i+1);
	    % Step 2: Compute Householder representation of V(:,i+1), i.e. 
	    % the submatrix V(i+1:n,i+1) will be transformed to a multiple
	    % of e_{i+1}
	    if i<n
	       [V(i+1:n,i+1),beta(i+1),MU]=HOUSE(V(i+1:n,i+1));
	       % Applying the Householder transformation to V(:,i+1) will
	       % give a multiple of the (i+1). unit vector. This multiple
	       % will be -sign(V_old(i+1,i+1))*MU=-sign(H(i+1,i))*MU.
	       if (H(i+1,i)>=0)
		  H(i+1,i)=-MU;
	       else   
		  H(i+1,i)=MU;
	       end % if-else
	    else	  
	       beta(i+1)=-2; H(i+1,i)=0;
	    end % if
else % CHOICE_ORTH==3
            % Compute modified Householder representation of V(:,i+1)
	    % Step 1: Multiply with previous Householder transformations,
	    %         V(:,i+1)= (I + V(:,1:i) T^{-*} V(:,1:i)')*Q
	    V(:,i+1)=Q+V(:,1:i)*(T(1:i,1:i)'\ (V(:,1:i)'*Q));
	    % => H(1:i,i)=V(:,1:i)'*Q=V(1:i,i+1)
	    H(1:min(n,i+1),i)=V(1:min(n,i+1),i+1);
	    % Step 2: Compute Householder representation of V(:,i+1), i.e. 
	    % the submatrix V(i+1:n,i+1) will be transformed to a multiple
	    % of e_{i+1}
	    if i<n
	       [V(i+1:n,i+1),beta(i+1),MU]=HOUSE(V(i+1:n,i+1));
	       V(1:i,i+1)=zeros(i,1);
	       % Applying the Householder transformation to V(:,i+1) will
	       % give a multiple of the (i+1). unit vector. This multiple
	       % will be -sign(V_old(i+1,i+1))*MU=-sign(H(i+1,i))*MU.
	       if (H(i+1,i)>=0)
		  H(i+1,i)=-MU;
	       else   
		  H(i+1,i)=MU;
	       end % if-else
	    else	  
	       beta(i+1)=-2; H(i+1,i)=0;
	    end % if
	    % Compute new column of T
	    T(i+1,i+1)=1/beta(i+1);
	    T(1:i,i+1)=-V(:,1:i)'*V(:,i+1);
	    % => HH_1*...*HH_{i+1} = I + V(:,1:i+1) T^{-1} V(:,1:i+1)'
end % -----------------------------------


            % Transform actual H to upper triangular form
	    for k=1:i-1
	        H(k:k+1,i)=ROWROT(H(k:k+1,i),c(k),s(k));
	    end % for
	    % Compute new Givens rotation
	    [c(i),s(i)]=GIVENS(H(i,i),H(i+1,i));
	    % Transform i. column   
	    H(i:i+1,i)=ROWROT(H(i:i+1,i),c(i),s(i));
	    % Transform rhs
	    rhs(i+1,1)=0;
	    rhs(i:i+1,1)=ROWROT(rhs(i:i+1,1),c(i),s(i));

	    % Will the transformed Hessenberg matrix become singular?
	    % nrmH(1:i+1)=nrmH(1:i+1)+abs(H(1:i+1,i));
	    % if abs(H(i,i))>n*eps*max(nrmH(1:i))
	       i=i+1;
	       kk=kk+1;
	    % else	  
	    %   nsgl=0;
	    %   if USE_SVD
	    %	  i=i+1;
	    %	  kk=kk+1;
	    %   end % if
	    % end % if-else	  
	    res(kk)=rhs(i);
		   	
		      
	    % if flag is set, compute the real residual
	    if stat	       
	       % test approximate solution
if USE_SVD % ----------------------------
               if nsgl 
		  y=H(1:i-1,1:i-1)\rhs(1:i-1,1);
	       else
		  y=pinv(H(1:i-1,1:i-1))*rhs(1:i-1,1);
	       end % if-else
else
	       y=H(1:i-1,1:i-1)\rhs(1:i-1,1);
end; % if USE_SVD -----------------------		  
if CHOICE_ORTH<2 % ----------------------
               xx=V(:,1:i-1)*y;
elseif CHOICE_ORTH==2	     
               xx=zeros(n,1);
	       xx(1:i-1)=y;
	       for k=i-1:-1:1
		   xx(k:n)=ROWHOUSE(xx(k:n),V(k:n,k),beta(k));
	       end % for
else % CHOICE_ORTH==3
               xx=zeros(n,1);
	       xx(1:i-1)=y;
	       xx=xx+V(:,1:i-1)*(T(1:i-1,1:i-1)\ (V(1:i-1,1:i-1)'*y));
end; % if CHOICE_ORTH<2 -----------------
               % ----------------------------------------------------
	       % Solve the preconditioned system M*xx=xx
	       if isstr(PARAM.P)
		  [xx,PARAM.PPARAM]=feval(PARAM.P,1,xx,x,PARAM.PPARAM);
	       else
		  xx=PARAM.P*xx+x;
	       end % if-else
	       % ----------------------------------------------------
	       % xx=x+xx;
	    end % if stat
		      
% print residual information	       
if PRINT_RES  % --------------------------
            fprintf('step %3d, ', kk);
	    fprintf('residual%16.8e', abs(res(kk))); 
	    semilogy(1:kk,abs(res(1:kk)),'-'); hold on;
	    % if flag is set, display the real residual
	    if stat	       
	       fprintf(', true residual%12.4e', norm(b-PARAM.A*xx));
	    end % if stat
	    fprintf('\n');
end % if PRINT_RES ---------------------
	     
      end % while
      % END of up to "restart" steps of GMRES

      % now compute approximate solution xsmall 
      if (i>1)
if USE_SVD % ----------------------------
         if nsgl 
	    y=H(1:i-1,1:i-1)\rhs(1:i-1,1);
	 else
	    y=pinv(H(1:i-1,1:i-1))*rhs(1:i-1,1);
	 end % if-else
else
	 y=triu(H(1:i-1,1:i-1))\rhs(1:i-1,1);
end % if USE_SVD -----------------------		  

% Select reorthogonalization scheme and corresponding computation of the
% next approximate solution
if CHOICE_ORTH<2 % ----------------------
	 % update x
	 xx=V(:,1:i-1)*y;
elseif CHOICE_ORTH==2
         xx=zeros(n,1);
	 xx(1:i-1)=y;
	 for k=i-1:-1:1
	     xx(k:n)=ROWHOUSE(xx(k:n),V(k:n,k),beta(k));
	 end % for
else % CHOICE_ORTH==3
         xx=zeros(n,1);
	 xx(1:i-1)=y;
	 % update x
	 xx=xx+V(:,1:i-1)*(T(1:i-1,1:i-1)\ (V(1:i-1,1:i-1)'*y));
end % if CHOICE_ORTH<2 -----------------
         % --------------------------------------------------------
	 % Solve the preconditioned system M*xx=xx
	 if isstr(PARAM.P)
	    [x,PARAM.PPARAM]=feval(PARAM.P,1,xx,x,PARAM.PPARAM);
	 else
	    x=PARAM.P*xx+x;
	 end % if-else
	 % ---------------------------------------------------------
	 % update x
	 % x=x+xx;
      end % if
end % while
PARAM.it=PARAM.it+kk;

	  
if PRINT_RES % -----------------
r=b-PARAM.A*x;
fprintf('2-norm of the final residual: %12.4e\n', norm(r,2));
end % if PRINT_RES % -----------


if stat % -----------------------
semilogy(1:length(res),abs(res),'-');
hold on
ylabel('norm of the residual');
title('history of the residual in the 2-norm');
end % if stat -------------------

if abs(res(kk))>PARAM.tol*abs(res(1)) & kk>PARAM.maxit
   fprintf('\n*****     Warning! GMRESJ did not converge!     ***\n\n');
end % if


PARAM.res=abs(res(1:kk));

if alpha==1
   Y=x+Y;
elseif alpha==-1
   Y=Y-x;
else
   Y=alpha*x+Y;
end % if-elseif
